@@ -59,7 +59,7 @@ std::unordered_set<Oomd::CgroupPath> resolveCgroupPaths(
     // (in resolveWildcardPath) every time is. If it's a big penalty,
     // we could search `absolutePath` for any fnmatch operators and
     // only resolve if we find symbols.
-    auto resolved_paths = Oomd::Fs::resolveWildcardPath(cgroup.absolutePath());
+    auto resolved_paths = Oomd::Fs::resolveWildcardPath(cgroup);
 
     for (const auto& resolved_path : resolved_paths) {
       size_t idx = 0;
@@ -169,7 +169,8 @@ int64_t Oomd::calculateProtection(
       return 0;
     }
 
-    return p_func(parent) * l_func(node->ctx) / l_sum_children;
+    return std::min(
+        l_func(node->ctx), p_func(parent) * l_func(node->ctx) / l_sum_children);
   };
 
   int64_t ret = p_func(node);
@@ -230,12 +231,14 @@ bool Oomd::updateContextRoot(const CgroupPath& path, OomdContext& ctx) {
   auto pressures = Fs::readMempressure("/");
   auto io_pressure = readIopressureWarnOnce("/");
   auto current = Fs::readMemcurrent("/");
+  auto nr_dying_descendants = Fs::getNrDyingDescendants(path.cgroupFs());
 
   ctx.setCgroupContext(
       path,
       {.pressure = pressures,
        .io_pressure = io_pressure,
-       .current_usage = current});
+       .current_usage = current,
+       .nr_dying_descendants = nr_dying_descendants});
 
   return true;
 }
@@ -262,25 +265,35 @@ bool Oomd::updateContextCgroup(const CgroupPath& path, OomdContext& ctx) {
   auto pressures = Fs::readMempressure(absolute_cgroup_path);
   auto memlow = Fs::readMemlow(absolute_cgroup_path);
   auto memmin = Fs::readMemmin(absolute_cgroup_path);
+  auto memhigh = Fs::readMemhigh(absolute_cgroup_path);
+  auto memmax = Fs::readMemmax(absolute_cgroup_path);
   auto swap_current = Fs::readSwapCurrent(absolute_cgroup_path);
   auto memory_stats = Fs::getMemstat(absolute_cgroup_path);
   auto anon_usage = memory_stats["anon"];
   auto io_pressure = readIopressureWarnOnce(absolute_cgroup_path);
+  int64_t memhigh_tmp = 0;
+  try {
+    memhigh_tmp = Fs::readMemhightmp(absolute_cgroup_path);
+  } catch (const Fs::bad_control_file& ex) {
+  }
   double io_cost_cumulative = 0;
   if (io_devs_.size() != 0) {
     IOStat io_stat;
-    try {
+    if (std::any_of(controllers.begin(), controllers.end(), [](std::string& s) {
+          return s == "io";
+        })) {
       io_stat = Fs::readIostat(absolute_cgroup_path);
-    } catch (const std::exception& ex) {
+    } else {
       if (!warned_io_stat_.count(absolute_cgroup_path)) {
         warned_io_stat_.emplace(absolute_cgroup_path);
-        OLOG << "IO stat unavailable on " << absolute_cgroup_path << ": "
-             << ex.what();
+        OLOG << "WARNING: cgroup io controller unavailable on "
+             << absolute_cgroup_path << ". io cost will be zero.";
       }
       io_stat = {};
     }
     io_cost_cumulative = calculateIOCostCumulative(io_stat);
   }
+  auto nr_dying_descendants = Fs::getNrDyingDescendants(absolute_cgroup_path);
 
   ctx.setCgroupContext(
       path,
@@ -291,7 +304,11 @@ bool Oomd::updateContextCgroup(const CgroupPath& path, OomdContext& ctx) {
        .swap_usage = swap_current,
        .anon_usage = anon_usage,
        .memory_min = memmin,
-       .io_cost_cumulative = io_cost_cumulative});
+       .memory_high = memhigh,
+       .memory_high_tmp = memhigh_tmp,
+       .memory_max = memmax,
+       .io_cost_cumulative = io_cost_cumulative,
+       .nr_dying_descendants = nr_dying_descendants});
 
   return true;
 }
@@ -381,21 +398,33 @@ void Oomd::updateContext(
   ctx = std::move(new_ctx);
 }
 
-int Oomd::prepDropInWatcher(const std::string& dir) {
+int Oomd::deregisterDropInWatcherFromEventLoop() {
   char buf[1024];
+  int ret = 0;
 
-  if (!Fs::isDir(dir)) {
-    OLOG << "Error: " << dir << " is not a directory";
-    return 1;
+  // Remove watch for stale fd and wd and reset them to their
+  // initial values. Caveat: If we try to remove the watch for
+  // drop_in_dir that has been deleted, inotify_rm_watch() returns
+  // EINVAL as the watch is automatically deleted when the directory
+  // was removed. Only remove watches if they're files in the directory
+  // and not the directory itself.
+  if (!drop_in_dir_deleted_ && ::inotify_rm_watch(inotifyfd_, inotifywd_) < 0) {
+    OLOG << "inotify_rm_watch: " << ::strerror_r(errno, buf, sizeof(buf));
+    ret = 1;
+  }
+  if (::epoll_ctl(epollfd_, EPOLL_CTL_DEL, inotifyfd_, nullptr) < 0) {
+    OLOG << "epoll_ctl: " << ::strerror_r(errno, buf, sizeof(buf));
+    ret = 1;
   }
 
-  // Load existing drop in configs before setting up inotify hooks
-  // so we don't race with configs dropping in
-  auto configs = Fs::readDir(dir, Fs::EntryType::REG_FILE);
-  std::sort(configs.begin(), configs.end()); // Provide some determinism
-  for (const auto& config : configs) {
-    processDropInAdd(config);
-  }
+  inotifyfd_ = -1;
+  inotifywd_ = -1;
+
+  return ret;
+}
+
+int Oomd::prepDropInWatcherEventLoop(const std::string& dir) {
+  char buf[1024];
 
   inotifyfd_ = ::inotify_init1(IN_NONBLOCK | IN_CLOEXEC);
   if (inotifyfd_ < 0) {
@@ -403,8 +432,9 @@ int Oomd::prepDropInWatcher(const std::string& dir) {
     return 1;
   }
 
-  uint32_t mask = IN_DELETE | IN_MODIFY | IN_MOVE | IN_ONLYDIR;
-  if (::inotify_add_watch(inotifyfd_, dir.c_str(), mask) < 0) {
+  uint32_t mask = IN_DELETE | IN_MODIFY | IN_MOVE | IN_ONLYDIR | IN_MOVE_SELF |
+      IN_DELETE_SELF;
+  if ((inotifywd_ = ::inotify_add_watch(inotifyfd_, dir.c_str(), mask)) < 0) {
     OLOG << "inotify_add_watch: " << ::strerror_r(errno, buf, sizeof(buf));
     return 1;
   }
@@ -422,6 +452,27 @@ int Oomd::prepDropInWatcher(const std::string& dir) {
   return 0;
 }
 
+int Oomd::prepDropInWatcher(const std::string& dir) {
+  if (!Fs::isDir(dir)) {
+    OLOG << "Error: " << dir << " is not a directory";
+    return 1;
+  }
+
+  // Load existing drop in configs before setting up inotify hooks
+  // so we don't race with configs dropping in
+  auto de = Fs::readDir(dir, Fs::DE_FILE);
+  std::sort(de.files.begin(), de.files.end()); // Provide some determinism
+  for (const auto& config : de.files) {
+    processDropInAdd(config);
+  }
+
+  if (prepDropInWatcherEventLoop(dir)) {
+    return 1;
+  }
+
+  return 0;
+}
+
 int Oomd::prepEventLoop(const std::chrono::seconds& interval) {
   char buf[1024];
 
@@ -540,7 +591,7 @@ int Oomd::processDropInWatcher(int fd) {
   while (true) {
     int len = ::read(fd, buf, sizeof(buf));
     if (len < 0 && errno != EAGAIN) {
-      OLOG << "epoll_wait: " << ::strerror_r(errno, buf, sizeof(buf));
+      OLOG << "read: " << ::strerror_r(errno, buf, sizeof(buf));
       return 1;
     }
 
@@ -560,6 +611,13 @@ int Oomd::processDropInWatcher(int fd) {
         // Remove drop in if file has been moved from or removed from
         // the watched directory
         processDropInRemove(event->name);
+      } else if (event->mask & (IN_DELETE_SELF | IN_MOVE_SELF)) {
+        // Remove stale watch descriptor for drop in if watched file or
+        // directory itself is moved or deleted
+        drop_in_dir_deleted_ = true;
+        if (deregisterDropInWatcherFromEventLoop()) {
+          return 1;
+        }
       }
     }
   }
@@ -590,6 +648,15 @@ int Oomd::processEventLoop() {
         OLOG << "read: " << ::strerror_r(errno, buf, sizeof(buf));
         return 1;
       }
+      // If drop_in_dir_ exists again and contains config files,
+      // set up the watch
+      if (drop_in_dir_deleted_ && Fs::isDir(drop_in_dir_)) {
+        ret = prepDropInWatcher(drop_in_dir_);
+        if (ret) {
+          return ret;
+        }
+        drop_in_dir_deleted_ = false;
+      }
     } else if (fd == inotifyfd_) {
       ret = processDropInWatcher(fd);
       if (ret) {
