@@ -40,7 +40,7 @@ class SystemdRestart : public Base {
   std::string service_;
   int post_action_delay_{15};
   bool dry_{false};
-  static constexpr auto kRestartsKey = "oomd.restarts_structured";
+  static constexpr auto kRestartsKey = "oomd.restarts";
 };
 
 } // namespace Oomd
