@@ -26,11 +26,12 @@
 #include <sys/xattr.h>
 #include <unistd.h>
 
+#include <cinttypes>
 #include <deque>
 #include <fstream>
+#include <sstream>
 #include <utility>
 
-#include "oomd/Log.h"
 #include "oomd/include/Assert.h"
 #include "oomd/util/Util.h"
 
@@ -62,14 +63,13 @@ PsiFormat getPsiFormat(const std::vector<std::string>& lines) {
 
 namespace Oomd {
 
-std::vector<std::string> Fs::readDir(const std::string& path, EntryType type) {
+struct Fs::DirEnts Fs::readDir(const std::string& path, int flags) {
   DIR* d;
-  std::vector<std::string> v;
+  struct Fs::DirEnts de;
 
   d = ::opendir(path.c_str());
   if (!d) {
-    OLOG << "Unable to open directory=" << path;
-    return v;
+    return de;
   }
 
   while (struct dirent* dir = ::readdir(d)) {
@@ -77,33 +77,38 @@ std::vector<std::string> Fs::readDir(const std::string& path, EntryType type) {
       continue;
     }
 
+    /*
+     * Optimisation: Avoid doing lstat calls if kernfs gives us back d_type.
+     * This actually can be pretty useful, since avoiding lstat()ing everything
+     * can reduce oomd CPU usage by ~10% on a reasonably sized cgroup
+     * hierarchy.
+     */
+    if ((flags & DirEntFlags::DE_FILE) && dir->d_type == DT_REG) {
+      de.files.push_back(dir->d_name);
+      continue;
+    }
+    if ((flags & DirEntFlags::DE_DIR) && dir->d_type == DT_DIR) {
+      de.dirs.push_back(dir->d_name);
+      continue;
+    }
+
     auto file = path + "/" + dir->d_name;
     struct stat buf;
     int ret = ::lstat(file.c_str(), &buf);
     if (ret == -1) {
-      auto errorNum = errno;
-      // TODO: better error handling for methods here instead of ignoring.
-      OLOG << "lstat failed with errno: " << errorNum << ", file: " << file
-           << ", ignoring file";
       continue;
     }
 
-    switch (type) {
-      case EntryType::REG_FILE:
-        if (buf.st_mode & S_IFREG) {
-          v.push_back(dir->d_name);
-        }
-        break;
-      case EntryType::DIRECTORY:
-        if (buf.st_mode & S_IFDIR) {
-          v.push_back(dir->d_name);
-        }
-        break;
+    if ((flags & DirEntFlags::DE_FILE) && (buf.st_mode & S_IFREG)) {
+      de.files.push_back(dir->d_name);
+    }
+    if ((flags & DirEntFlags::DE_DIR) && (buf.st_mode & S_IFDIR)) {
+      de.files.push_back(dir->d_name);
     }
   }
 
   ::closedir(d);
-  return v;
+  return de;
 }
 
 bool Fs::isDir(const std::string& path) {
@@ -115,18 +120,32 @@ bool Fs::isDir(const std::string& path) {
   return false;
 }
 
+/*
+ * Return if a string might have something special for fnmatch.
+ *
+ * This function is simple and can return false-positives, but not
+ * false-negatives -- that is, true means "maybe", and false means false.
+ * That's ok, since this is only used for optimisations.
+ */
+bool Fs::hasGlob(const std::string& s) {
+  return s.find_first_of("*[?") != std::string::npos;
+}
+
 std::unordered_set<std::string> Fs::resolveWildcardPath(
-    const std::string& path) {
+    const CgroupPath& cgpath) {
+  std::string path = cgpath.absolutePath();
   std::unordered_set<std::string> ret;
   if (path.empty()) {
     return ret;
   }
 
   auto parts = Util::split(path, '/');
+  OCHECK_EXCEPT(!parts.empty(), "No parts in " + path);
+
   std::deque<std::pair<std::string, size_t>> queue;
-  // Add initial path piece to begin search on. Start at root
-  // if provided path is absolute, else go with relative dir.
-  queue.emplace_back((path[0] == '/' ? "/" : "./"), 0);
+
+  // Add initial path piece to begin search on. Start at root.
+  queue.emplace_back("/", 0);
 
   // Perform a DFS on the entire search space. Note that we pattern
   // match at each level of the provided path to eliminate "dead"
@@ -140,17 +159,25 @@ std::unordered_set<std::string> Fs::resolveWildcardPath(
     const auto front = queue.front(); // copy
     queue.pop_front();
 
+    // Optimisation: If there's no glob and we're not at the end, it must be
+    // intended to be a single dir. It doesn't matter if it actually *is* in
+    // reality, because if it doesn't exist we'll fail later on.
+    if (front.second < parts.size() - 1 && !Fs::hasGlob(parts[front.second])) {
+      queue.emplace_front(
+          front.first + parts[front.second] + "/", front.second + 1);
+      continue;
+    }
+
     // We can't continue BFS if we've hit a regular file
     if (!isDir(front.first)) {
       continue;
     }
 
-    // Only resolve regular files and directories
-    auto entries = readDir(front.first, EntryType::DIRECTORY);
-    auto files = readDir(front.first, EntryType::REG_FILE);
-    entries.insert(entries.end(), files.begin(), files.end());
+    auto de = readDir(front.first, DE_FILE | DE_DIR);
+    de.files.reserve(de.files.size() + de.dirs.size());
+    de.files.insert(de.files.end(), de.dirs.begin(), de.dirs.end());
 
-    for (const auto& entry : entries) {
+    for (const auto& entry : de.files) {
       if (::fnmatch(parts[front.second].c_str(), entry.c_str(), 0) == 0) {
         if (front.second == parts.size() - 1) {
           // We have reached a leaf, add it to the return set
@@ -168,42 +195,6 @@ std::unordered_set<std::string> Fs::resolveWildcardPath(
     }
   }
 
-  // Clean up paths a little
-  //
-  // Note we can't do an in-place modification b/c std::unordered_set needs
-  // to hash elements and disallows any modifications to while iterating. We
-  // must instead batch deletes and inserts.
-  std::unordered_set<std::string> to_remove;
-  std::unordered_set<std::string> to_add;
-  for (const auto& p : ret) {
-    // Remove leading "./"
-    if (p.size() >= 2 && p.at(0) == '.' && p.at(1) == '/') {
-      to_remove.emplace(p);
-      to_add.emplace(p.substr(2));
-    }
-  }
-  for (const auto& p : to_remove) {
-    ret.erase(p);
-  }
-  ret.insert(to_add.begin(), to_add.end());
-
-  return ret;
-}
-
-std::unordered_set<CgroupPath> Fs::resolveCgroupWildcardPath(
-    const CgroupPath& path) {
-  std::unordered_set<CgroupPath> ret;
-  auto resolved_raw_paths = resolveWildcardPath(path.absolutePath());
-  for (const auto& raw : resolved_raw_paths) {
-    // The fully resolved path being shorter than the cgroup fs path
-    // should never really happen but we error check anyways
-    if (raw.size() < path.cgroupFs().size()) {
-      continue;
-    }
-
-    ret.emplace(path.cgroupFs(), raw.substr(path.cgroupFs().size()));
-  }
-
   return ret;
 }
 
@@ -223,7 +214,6 @@ void Fs::removePrefix(std::string& str, const std::string& prefix) {
 std::vector<std::string> Fs::readFileByLine(const std::string& path) {
   std::ifstream f(path, std::ios::in);
   if (!f.is_open()) {
-    OLOG << "Unable to open " << path;
     return {};
   }
 
@@ -233,10 +223,6 @@ std::vector<std::string> Fs::readFileByLine(const std::string& path) {
     v.push_back(std::move(s));
   }
 
-  if (f.bad()) {
-    OLOG << "Error while processing file " << path;
-  }
-
   return v;
 }
 
@@ -254,20 +240,18 @@ std::vector<std::string> Fs::readControllers(const std::string& path) {
 
 std::vector<int> Fs::getPids(const std::string& path, bool recursive) {
   std::vector<int> pids;
-  auto files = readDir(path, EntryType::REG_FILE);
-  if (std::any_of(files.begin(), files.end(), [](const std::string& s) {
+  auto de = readDir(path, DE_FILE | DE_DIR);
+  if (std::any_of(de.files.begin(), de.files.end(), [](const std::string& s) {
         return s == kProcsFile;
       })) {
     auto str_pids = readFileByLine(path + "/" + kProcsFile);
     for (const auto& sp : str_pids) {
-      OLOG << "found pid " << sp;
       pids.push_back(std::stoi(sp));
     }
   }
 
   if (recursive) {
-    auto dirs = readDir(path, EntryType::DIRECTORY);
-    for (const auto& dir : dirs) {
+    for (const auto& dir : de.dirs) {
       auto recursive_pids = getPids(path + "/" + dir, true);
       pids.insert(pids.end(), recursive_pids.begin(), recursive_pids.end());
     }
@@ -347,6 +331,7 @@ ResourcePressure Fs::readRespressure(
           std::stof(toks[1]),
           std::stof(toks[2]),
           std::stof(toks[3]),
+          std::nullopt,
       };
     }
     case PsiFormat::MISSING:
@@ -404,6 +389,10 @@ int64_t Fs::readMemhigh(const std::string& path) {
   return Fs::readMinMaxLowHigh(path, kMemHighFile);
 }
 
+int64_t Fs::readMemmax(const std::string& path) {
+  return Fs::readMinMaxLowHigh(path, kMemMaxFile);
+}
+
 int64_t Fs::readMemhightmp(const std::string& path) {
   auto lines = readFileByLine(path + "/" + kMemHighTmpFile);
   OCHECK_EXCEPT(lines.size() == 1, bad_control_file(path + ": missing file"));
@@ -456,12 +445,13 @@ std::unordered_map<std::string, int64_t> Fs::getVmstat(
 std::unordered_map<std::string, int64_t> Fs::getMeminfo(
     const std::string& path) {
   char name[256] = {0};
-  unsigned long val;
+  uint64_t val;
   std::unordered_map<std::string, int64_t> map;
 
   auto lines = readFileByLine(path);
   for (auto& line : lines) {
-    int ret = sscanf(line.c_str(), "%255[^:]:%*[ \t]%lu%*s\n", name, &val);
+    int ret =
+        sscanf(line.c_str(), "%255[^:]:%*[ \t]%" SCNu64 "%*s\n", name, &val);
     if (ret == 2) {
       map[name] = val * 1024;
     }
@@ -470,15 +460,15 @@ std::unordered_map<std::string, int64_t> Fs::getMeminfo(
   return map;
 }
 
-std::unordered_map<std::string, int64_t> Fs::getMemstat(
-    const std::string& path) {
+std::unordered_map<std::string, int64_t> Fs::getMemstatLike(
+    const std::string& file) {
   char name[256] = {0};
-  unsigned long val;
+  uint64_t val;
   std::unordered_map<std::string, int64_t> map;
 
-  auto lines = readFileByLine(path + "/" + kMemStatFile);
+  auto lines = readFileByLine(file);
   for (const auto& line : lines) {
-    int ret = sscanf(line.c_str(), "%255s %lu\n", name, &val);
+    int ret = sscanf(line.c_str(), "%255s %" SCNu64 "\n", name, &val);
     if (ret == 2) {
       map[name] = val;
     }
@@ -487,6 +477,11 @@ std::unordered_map<std::string, int64_t> Fs::getMemstat(
   return map;
 }
 
+std::unordered_map<std::string, int64_t> Fs::getMemstat(
+    const std::string& path) {
+  return getMemstatLike(path + "/" + kMemStatFile);
+}
+
 ResourcePressure Fs::readIopressure(
     const std::string& path,
     PressureType type) {
@@ -511,7 +506,8 @@ IOStat Fs::readIostat(const std::string& path) {
     int major, minor;
     int ret = sscanf(
         line.c_str(),
-        "%d:%d rbytes=%lu wbytes=%lu rios=%lu wios=%lu dbytes=%lu dios=%lu\n",
+        "%d:%d rbytes=%" SCNd64 " wbytes=%" SCNd64 " rios=%" SCNd64
+        " wios=%" SCNd64 " dbytes=%" SCNd64 " dios=%" SCNd64 "\n",
         &major,
         &minor,
         &dev_io_stat.rbytes,
@@ -567,14 +563,18 @@ void Fs::writeMemhightmp(
   }
 }
 
+int64_t Fs::getNrDyingDescendants(const std::string& path) {
+  auto map = getMemstatLike(path + "/" + kCgroupStatFile);
+  // Will return 0 for missing entries
+  return map["nr_dying_descendants"];
+}
+
 bool Fs::setxattr(
     const std::string& path,
     const std::string& attr,
     const std::string& val) {
   int ret = ::setxattr(path.c_str(), attr.c_str(), val.c_str(), val.size(), 0);
   if (ret == -1) {
-    OLOG << "Unable to set xattr " << attr << "=" << val << " on " << path
-         << ". errno=" << errno;
     return false;
   }
   return true;
